# Copyright (c) 2016, 2017, 2018 Forschungszentrum Juelich GmbH
# Author: Yann Leprince <y.leprince@fz-juelich.de>
# 
# Minor later modifications (c) 2018 by William Silversmith <ws9@princeton.edu>
#
# This software is made available under the MIT license, see bottom of file.

from six.moves import zip_longest
import functools
import itertools
import struct
import sys

import numpy as np

def ceil_div(a, b):
    """Ceil integer division (``ceil(a / b)`` using integer arithmetic)."""
    return (a - 1) // b + 1


class InvalidFormatError(Exception):
    """Raised when chunk data cannot be decoded properly."""
    pass

def pad_block(block, block_size):
    """Pad a block to block_size with its most frequent value"""
    unique_vals, unique_counts = np.unique(block, return_counts=True)
    most_frequent_value = unique_vals[np.argmax(unique_counts)]
    return np.pad(block,
                  tuple((0, desired_size - actual_size)
                        for desired_size, actual_size
                        in zip(block_size, block.shape)),
                  mode="constant", constant_values=most_frequent_value)


def number_of_encoding_bits(elements):
    for bits in (0, 1, 2, 4, 8, 16, 32):
        if 2 ** bits >= elements:
            return bits
    raise ValueError("Too many elements!")


COMPRESSED_SEGMENTATION_DATA_TYPES = (
    np.dtype(np.uint32).newbyteorder("<"),
    np.dtype(np.uint64).newbyteorder("<"),
)


def encode_chunk(chunk, block_size):
    # Construct file in memory step by step
    num_channels = chunk.shape[0]
    buf = bytearray(4 * num_channels)

    assert chunk.dtype in COMPRESSED_SEGMENTATION_DATA_TYPES

    for channel in range(num_channels):
        # Write offset of the current channel into the header
        assert len(buf) % 4 == 0
        struct.pack_into("<I", buf, channel * 4, len(buf) // 4)

        buf += _encode_channel(
            chunk[channel, :, :, :], block_size)
    return buf


def _encode_channel(chunk_channel, block_size):
    # Grid size (number of blocks in the chunk)
    gx = ceil_div(chunk_channel.shape[2], block_size[0])
    gy = ceil_div(chunk_channel.shape[1], block_size[1])
    gz = ceil_div(chunk_channel.shape[0], block_size[2])
    stored_lut_offsets = {}
    buf = bytearray(gx * gy * gz * 8)
    for z, y, x in np.ndindex((gz, gy, gx)):
        block = chunk_channel[
            z*block_size[2] : (z+1)*block_size[2],
            y*block_size[1] : (y+1)*block_size[1],
            x*block_size[0] : (x+1)*block_size[0]
        ]
        if block.shape != block_size:
            block = pad_block(block, block_size)

        # TODO optimization: to improve additional compression (gzip), sort the
        # list of unique symbols by decreasing frequency using
        # return_counts=True so that low-value symbols are used more often.
        # Alternatively, sort by label value to improve sharing of lookup
        # tables...
        (lookup_table, encoded_values) = np.unique(
            block, return_inverse=True, return_counts=False)
        bits = number_of_encoding_bits(len(lookup_table))

        # Write look-up table to the buffer (or re-use another one)
        lut_bytes = lookup_table.astype(block.dtype).tobytes()
        if lut_bytes in stored_lut_offsets:
            lookup_table_offset = stored_lut_offsets[lut_bytes]
        else:
            assert len(buf) % 4 == 0
            lookup_table_offset = len(buf) // 4
            buf += lut_bytes
            stored_lut_offsets[lut_bytes] = lookup_table_offset

        assert len(buf) % 4 == 0
        encoded_values_offset = len(buf) // 4
        buf += _pack_encoded_values(encoded_values, bits)

        assert lookup_table_offset == (lookup_table_offset & 0xFFFFFF)
        struct.pack_into("<II", buf, 8 * (x + gx * (y + gy * z)),
                         lookup_table_offset | (bits << 24),
                         encoded_values_offset)
    return buf


def _pack_encoded_values(encoded_values, bits):
    # TODO optimize with np.packbits for bits == 1
    if bits == 0:
        return bytes()
    else:
        values_per_32bit = 32 // bits
        assert np.all(encoded_values == encoded_values & ((1 << bits) - 1))
        padded_values = np.empty(
            values_per_32bit * ceil_div(len(encoded_values), values_per_32bit),
            dtype="<I")
        padded_values[:len(encoded_values)] = encoded_values
        padded_values[len(encoded_values):] = 0
        packed_values = functools.reduce(
            np.bitwise_or,
            (padded_values[shift::values_per_32bit] << (shift * bits)
             for shift in range(values_per_32bit)))
        return packed_values.tobytes()


def decode_chunk_into(chunk, buf, block_size):
    num_channels = chunk.shape[0]
    # Grid size (number of blocks in the chunk)
    gx = ceil_div(chunk.shape[3], block_size[0])
    gy = ceil_div(chunk.shape[2], block_size[1])
    gz = ceil_div(chunk.shape[1], block_size[2])

    if len(buf) < num_channels * (4 + 8 * gx * gy * gz):
        raise InvalidFormatError("compressed_segmentation file too short")

    if sys.version_info < (3,):
        channel_offsets = struct.unpack("<I", buf[:4*num_channels])
        channel_offsets = [ 4 * ret for ret in channel_offsets ]
    else:
        channel_offsets = [
            4 * ret[0] for ret in struct.iter_unpack("<I", buf[:4*num_channels])
        ]

    for channel, (offset, next_offset) in \
        enumerate(zip_longest(channel_offsets, channel_offsets[1:])):

        # next_offset will be None for the last channel
        if offset + 8 * gx * gy * gz > len(buf):
            raise InvalidFormatError("compressed_segmentation channel offset "
                                     "is too large (truncated file?)")
        _decode_channel_into(
            chunk, channel, buf[offset:next_offset], block_size
        )

    return chunk


def _decode_channel_into(chunk, channel, buf, block_size):
    # Grid size (number of blocks in the chunk)
    gx = ceil_div(chunk.shape[3], block_size[0])
    gy = ceil_div(chunk.shape[2], block_size[1])
    gz = ceil_div(chunk.shape[1], block_size[2])
    block_num_elem = block_size[0] * block_size[1] * block_size[2]
    for z, y, x in np.ndindex((gz, gy, gx)):
        # Read the block header
        res = struct.unpack_from("<II", buf, 8 * (x + gx * (y + gy * z)))
        lookup_table_offset = 4 * (res[0] & 0x00FFFFFF)
        bits = res[0] >> 24
        if bits not in (0, 1, 2, 4, 8, 16, 32):
            raise InvalidFormatError("Invalid number of encoding bits for "
                                     "compressed_segmentation block ({0})"
                                     .format(bits))
        encoded_values_offset = 4 * res[1]
        if bits != 0 and encoded_values_offset > len(buf):
            raise InvalidFormatError("Invalid offset for encoded values in "
                                     "compressed_segmentation block "
                                     "(truncated file?)")
        lookup_table_past_end = lookup_table_offset + chunk.itemsize * min(
            (2 ** bits),
            ((len(buf) - lookup_table_offset) // chunk.itemsize)
        )
        lookup_table = np.frombuffer(
            buf[lookup_table_offset:lookup_table_past_end], dtype=chunk.dtype)
        if bits == 0:
            block = np.empty(block_size, dtype=chunk.dtype)
            try:
                block[...] = lookup_table[0]
            except IndexError as exc:
                raise InvalidFormatError(
                    "Invalid compressed_segmentation data: indexing out of "
                    "the lookup table")
        else:
            values_per_32bit = 32 // bits
            encoded_values_end = encoded_values_offset + 4 * (
                ceil_div(block_num_elem, values_per_32bit)
            )
            packed_values = np.frombuffer(buf[encoded_values_offset:
                                              encoded_values_end], dtype="<I")
            encoded_values = _unpack_encoded_values(packed_values, bits,
                                                    block_num_elem)
            # Apply the lookup table
            try:
                decoded_values = lookup_table[encoded_values]
            except IndexError as exc:
                raise InvalidFormatError(
                    "Invalid compressed_segmentation data: indexing out of "
                    "the lookup table")
            block = decoded_values.reshape((block_size[2],
                                            block_size[1],
                                            block_size[0]))

        # Remove padding
        zmax = min(block_size[2], chunk.shape[1] - z * block_size[2])
        ymax = min(block_size[1], chunk.shape[2] - y * block_size[1])
        xmax = min(block_size[0], chunk.shape[3] - x * block_size[0])
        chunk[
            channel,
            z*block_size[2] : (z+1)*block_size[2],
            y*block_size[1] : (y+1)*block_size[1],
            x*block_size[0] : (x+1)*block_size[0]
        ] = block[:zmax, :ymax, :xmax]


def _unpack_encoded_values(packed_values, bits, num_values):
    if bits == 0:
        return np.zeros(num_values, dtype="<I")
    else:
        bitmask = (1 << bits) - 1
        values_per_32bit = 32 // bits
        padded_values = np.empty(
            values_per_32bit * ceil_div(num_values, values_per_32bit),
            dtype="<I")
        for shift in range(values_per_32bit):
            padded_values[shift::values_per_32bit] = (
                (packed_values >> (shift * bits)) & bitmask)
        return padded_values[:num_values]


# MIT License

# Copyright (c) 2016 Forschungszentrum Juelich GmbH

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import concurrent.futures
import copy
from functools import partial
import itertools
import json
import math
import multiprocessing as mp
import os
import posixpath
import signal

import numpy as np

from ....lib import xyzrange, min2, max2, Vec, Bbox
from .... import sharedmemory as shm

# Used in sharedmemory to emulate shared memory on 
# OS X using a file, which has that facility but is 
# more limited than on Linux.
fs_lock = mp.Lock()

def parallel_execution(fn, items, parallel, cleanup_shm=None):
  def cleanup(signum, frame):
    if cleanup_shm:
      shm.unlink(cleanup_shm)

  prevsigint = signal.getsignal(signal.SIGINT)
  prevsigterm = signal.getsignal(signal.SIGTERM)

  signal.signal(signal.SIGINT, cleanup)
  signal.signal(signal.SIGTERM, cleanup)

  with concurrent.futures.ProcessPoolExecutor(max_workers=parallel) as executor:
    executor.map(fn, items)

  signal.signal(signal.SIGINT, prevsigint)
  signal.signal(signal.SIGTERM, prevsigterm)

def chunknames(bbox, volume_bbox, key, chunk_size, protocol=None):
  path = posixpath if protocol != 'file' else os.path

  for x,y,z in xyzrange( bbox.minpt, bbox.maxpt, chunk_size ):
    highpt = min2(Vec(x,y,z) + chunk_size, volume_bbox.maxpt)
    filename = "{}-{}_{}-{}_{}-{}".format(
      x, highpt.x,
      y, highpt.y, 
      z, highpt.z
    )
    yield path.join(key, filename)

def gridpoints(bbox, volume_bbox, chunk_size):
  """
  Consider a volume as divided into a grid with the 
  first chunk labeled 1, the second 2, etc. 

  Return the grid x,y,z coordinates of a cutout as a
  sequence.
  """
  chunk_size = Vec(*chunk_size)

  grid_size = np.ceil(volume_bbox.size3() / chunk_size).astype(np.int64)
  cutout_grid_size = np.ceil(bbox.size3() / chunk_size).astype(np.int64)
  cutout_grid_offset = np.ceil((bbox.minpt - volume_bbox.minpt) / chunk_size).astype(np.int64)

  grid_cutout = Bbox( cutout_grid_offset, cutout_grid_offset + cutout_grid_size )

  for x,y,z in xyzrange( grid_cutout.minpt, grid_cutout.maxpt, (1,1,1) ):
    yield Vec(x,y,z)

def compressed_morton_code(gridpt, grid_size):
  gridpt = np.asarray(gridpt, dtype=np.uint32)
  single_input = False
  if gridpt.ndim == 1:
    gridpt = np.atleast_2d(gridpt)
    single_input = True

  code = np.zeros((gridpt.shape[0],), dtype=np.uint64)
  num_bits = max(( math.ceil(math.log2(size)) for size in grid_size ))
  j = np.uint64(0)
  one = np.uint64(1)

  for i in range(num_bits):
    for dim in range(3):
      if 2 ** i <= grid_size[dim]:
        bit = (((np.uint64(gridpt[:, dim]) >> np.uint64(i)) & one) << j)
        code |= bit
        j += one

  if single_input:
    return code[0]
  return code

def shade(dest_img, dest_bbox, src_img, src_bbox):
  """
  Shade dest_img at coordinates dest_bbox using the
  image contained in src_img at coordinates src_bbox.

  The buffer will only be painted in the overlapping
  region of the content.

  Returns: void
  """
  if not Bbox.intersects(dest_bbox, src_bbox):
    return

  spt = max2(src_bbox.minpt, dest_bbox.minpt)
  ept = min2(src_bbox.maxpt, dest_bbox.maxpt)
  dbox = Bbox(spt, ept) - dest_bbox.minpt

  ZERO3 = Vec(0,0,0)
  istart = max2(spt - src_bbox.minpt, ZERO3)
  iend = min2(ept - src_bbox.maxpt, ZERO3) + src_img.shape[:3]
  sbox = Bbox(istart, iend)

  while src_img.ndim < 4:
    src_img = src_img[..., np.newaxis]
  
  dest_img[ dbox.to_slices() ] = src_img[ sbox.to_slices() ]




import six
from six.moves import queue as Queue
from collections import defaultdict
import itertools
import json
import os.path
import posixpath
import re
from functools import partial
import types

from tqdm import tqdm

from cloudvolume import compression
from cloudvolume.exceptions import UnsupportedProtocolError
from cloudvolume.lib import mkdir, scatter, jsonify, duplicates
from cloudvolume.threaded_queue import ThreadedQueue, DEFAULT_THREADS
from cloudvolume.scheduler import schedule_green_jobs

import cloudvolume.paths

from .storage_interfaces import (
  FileInterface, HttpInterface, 
  S3Interface, GoogleCloudStorageInterface
)

def get_interface_class(protocol):
  if protocol == 'file':
    return FileInterface
  elif protocol == 'gs':
    return GoogleCloudStorageInterface
  elif protocol in ('s3', 'matrix'):
    return S3Interface
  elif protocol in ('http', 'https'):
    return HttpInterface
  else:
    raise UnsupportedProtocolError(str(self._path))

def default_byte_iterator(starts, ends):
  if starts is None:
    starts = itertools.repeat(None)
  if ends is None:
    ends = itertools.repeat(None)
  return iter(starts), iter(ends)

class StorageBase(object):
  """Abastract base class of Storage with some implementation details."""
  def __init__(self, layer_path, progress=False):
    self.progress = progress

    self._layer_path = layer_path
    self._path = cloudvolume.paths.extract(layer_path)
    self._interface_cls = get_interface_class(self._path.protocol)
  
  @property
  def layer_path(self):
    return self._layer_path

  def progress_description(self, prefix):
    if isinstance(self.progress, str):
      return prefix + ' ' + self.progress
    else:
      return prefix if self.progress else None

  def get_connection(self):
    return self._interface_cls(self._path)

  def get_path_to_file(self, file_path):
    return posixpath.join(self._layer_path, file_path)

  def put_json(self, file_path, content, content_type='application/json', *args, **kwargs):
    if type(content) != str:
      content = jsonify(content)
    return self.put_file(file_path, content, content_type=content_type, *args, **kwargs)
    
  def get_json(self, file_path):
    content = self.get_file(file_path)
    if content is None:
      return None
    return json.loads(content.decode('utf8'))

  def put_file(self, file_path, content, content_type=None, compress=None, compress_level=None, cache_control=None):
    """ 
    Args:
      filename (string): it can contains folders
      content (string): binary data to save
    """
    return self.put_files([ (file_path, content) ], 
      content_type=content_type, 
      compress=compress, 
      compress_level=compress_level,
      cache_control=cache_control, 
      block=False
    )

  def exists(self, file_path):
    raise NotImplementedError()

  def files_exist(self, file_paths):
    raise NotImplementedError()

  def get_file(self, file_path):
    raise NotImplementedError()

  def get_files(self, file_paths, starts=None, ends=None):
    raise NotImplementedError()

  def delete_file(self, file_path):
    raise NotImplementedError()

  def delete_files(self, file_paths):
    raise NotImplementedError()

  def list_files(self, prefix="", flat=False):
    raise NotImplementedError()

  def __del__(self):
    pass

  def __enter__(self):
    return self

  def __exit__(self, exception_type, exception_value, traceback):
    pass

class SimpleStorage(StorageBase):
  """
  Access files stored in Google Storage (gs), Amazon S3 (s3), 
  or the local Filesystem (file).

  e.g. with Storage('gs://bucket/dataset/layer') as stor:
      files = stor.get_file('filename')

  Required:
    layer_path (str): A protocol prefixed path of the above format.
      Accepts s3:// gs:// and file://. File paths are absolute.

  Optional:
    progress (bool:false): Show a tqdm progress bar for multiple 
      uploads and downloads.
  """
  def __init__(self, layer_path, progress=False):
    super(SimpleStorage, self).__init__(layer_path, progress)
    self._interface = self.get_connection()

  def put_files(self, files, content_type=None, compress=None, compress_level=None, cache_control=None, block=True):
    """
    Put lots of files at once and get a nice progress bar. It'll also wait
    for the upload to complete, just like get_files.

    Required:
      files: [ (filepath, content), .... ]
    """
    desc = self.progress_description('Uploading')
    for path, content in tqdm(files, disable=(not self.progress), desc=desc):
      content = compression.compress(content, method=compress, compress_level=compress_level)
      self._interface.put_file(path, content, content_type, compress, cache_control=cache_control)
    return self

  def exists(self, file_path):
    """Test if a single file exists. Returns boolean."""
    return self._interface.exists(file_path)

  def files_exist(self, file_paths):
    """
    Threaded exists for all file paths. 

    file_paths: (list) file paths to test for existence

    Returns: { filepath: bool }
    """
    return self._interface.files_exist(file_paths)

  def get_file(self, file_path, start=None, end=None):
    """
    Get the binary contents of a file. Optionally, specify
    the inclusive byte range to request.
    """
    content, encoding = self._interface.get_file(file_path, start=start, end=end)
    content = compression.decompress(content, encoding, filename=file_path)
    return content

  def get_files(self, file_paths, starts=None, ends=None):
    starts, ends = default_byte_iterator(starts, ends)

    iterator = tqdm(
      zip(file_paths, starts, ends),
      disable=(not self.progress), 
      desc=self.progress_description("Downloading")
    )

    results = []
    for path, start, end in iterator:
      error = None 

      try:
        content = self.get_file(path, start, end)
      except Exception as err:
        error = err 
        content = None 

      results.append({
        'filename': path,
        'byte_range': (start, end),
        'content': content,
        'error': error,
      })

    return results 

  def delete_file(self, file_path):
    self._interface.delete_file(file_path)

  def delete_files(self, file_paths):
    for path in file_paths:
      self._interface.delete_file(path)
    return self

  def list_files(self, prefix="", flat=False):
    """
    List the files in the layer with the given prefix. 

    flat means only generate one level of a directory,
    while non-flat means generate all file paths with that 
    prefix.

    Here's how flat=True handles different senarios:
      1. partial directory name prefix = 'bigarr'
        - lists the '' directory and filters on key 'bigarr'
      2. full directory name prefix = 'bigarray'
        - Same as (1), but using key 'bigarray'
      3. full directory name + "/" prefix = 'bigarray/'
        - Lists the 'bigarray' directory
      4. partial file name prefix = 'bigarray/chunk_'
        - Lists the 'bigarray/' directory and filters on 'chunk_'
    
    Return: generated sequence of file paths relative to layer_path
    """

    for f in self._interface.list_files(prefix, flat):
      yield f

  def __del__(self):
    self._interface.release_connection()

  def __enter__(self):
    return self

  def __exit__(self, exception_type, exception_value, traceback):
    self._interface.release_connection()

  def __getitem__(self, key):
    return self.get_file(key)
  
  def __setitem__(self, key, value):
    if type(key) == tuple:
      key, kwargs = key
    else:
      kwargs = {}

    self.put_item(key, value, **kwargs)

class GreenStorage(StorageBase):
  def __init__(
    self, layer_path, progress=False, 
    n_threads=DEFAULT_THREADS, 
  ):
    import gevent.pool
    super(GreenStorage, self).__init__(layer_path, progress)
    self.concurrency = n_threads if n_threads > 0 else 1
    self.pool = gevent.pool.Pool(self.concurrency)

  def exists(self, file_path):
    """Test if a single file exists. Returns boolean."""
    with self.get_connection() as conn:
      return conn.exists(file_path)

  def files_exist(self, file_paths):
    """
    Threaded exists for all file paths.

    file_paths: (list) file paths to test for existence

    Returns: { filepath: bool }
    """
    results = {}

    def exist_thunk(paths):
      with self.get_connection() as conn:
        results.update(conn.files_exist(paths))
    
    desc = self.progress_description('Existence Testing')
    schedule_green_jobs(  
      fns=( partial(exist_thunk, paths) for paths in scatter(file_paths, self.concurrency) ),
      progress=(desc if self.progress else None),
      concurrency=self.concurrency,
      total=len(file_paths),
    )

    return results

  def get_file(self, file_path, start=None, end=None):
    """
    Get the binary contents of a file. Optionally, specify
    the inclusive byte range to request.
    """
    with self.get_connection() as conn:
      content, encoding = conn.get_file(file_path, start=start, end=end)
    return compression.decompress(content, encoding, filename=file_path)

  def get_files(self, file_paths, starts=None, ends=None):
    """
    Returns: [ 
      { "filename": ..., "content": bytes, "error": exception or None }, 
      ... 
    ]
    """
    starts, ends = default_byte_iterator(starts, ends)

    def getfn(path, start, end):
      result = error = None 

      conn = self.get_connection()
      try:
        result = conn.get_file(path, start=start, end=end)
      except Exception as err:
        error = err
        # important to print immediately because 
        # errors are collected at the end
        print(err)
        del conn
      else:
        conn.release_connection()
      
      content, encoding = result
      content = compression.decompress(content, encoding)

      return {
        "filename": path,
        "byte_range": (start, end),
        "content": content,
        "error": error,
      }

    desc = self.progress_description('Downloading')

    return schedule_green_jobs(  
      fns=( 
        partial(getfn, path, start, end) 
        for path, start, end in zip(file_paths, starts, ends) 
      ),
      progress=(desc if self.progress else None),
      concurrency=self.concurrency,
      total=len(file_paths),
    )

  def put_files(
    self, files, 
    content_type=None, compress=None, compress_level=None,
    cache_control=None, block=True
  ):
    """
    Put lots of files at once and get a nice progress bar. It'll also wait
    for the upload to complete, just like get_files.

    Required:
      files: [ (filepath, content), .... ]
    """ 
    if compress not in compression.COMPRESSION_TYPES:
      raise NotImplementedError()

    def uploadfn(path, content):
      with self.get_connection() as conn:
        content = compression.compress(content, method=compress, compress_level=compress_level)
        conn.put_file(
          file_path=path, 
          content=content, 
          content_type=content_type, 
          compress=compress, 
          cache_control=cache_control,
        )

    if not isinstance(gen, types.GeneratorType):
      dupes = duplicates([ path for path, content in files ])
      if dupes:
        raise ValueError("Cannot write the same file multiple times in one pass. This causes a race condition. Files: " + ", ".join(dupes))

    fns = ( partial(uploadfn, path, content) for path, content in files )

    if block:
      desc = desc = self.progress_description('Uploading')
      schedule_green_jobs(
        fns=fns,
        progress=(desc if self.progress else None),
        concurrency=self.concurrency,
        total=len(files),
      )
    else:
      for fn in fns:
        self.pool.spawn(fn)

    return self

  def wait(self, desc=None):
    self.pool.join()

  def start_threads(self):
    import gevent.pool
    self.pool.kill()
    self.pool = gevent.pool.Pool(self.concurrency)

  def kill_threads(self):
    self.pool.kill()

  def delete_file(self, file_path):
    with self.get_connection() as conn:
      conn.delete_file(file_path)
    return self

  def delete_files(self, file_paths):
    def thunk_delete(path):
      with self.get_connection() as conn:
        conn.delete_file(path)

    desc = self.progress_description('Deleting')

    schedule_green_jobs(
      fns=( partial(thunk_delete, path) for path in file_paths ),
      progress=(desc if self.progress else None),
      concurrency=self.concurrency,
      total=len(file_paths),
    )
    
    return self

  def list_files(self, prefix="", flat=False):
    """
    List the files in the layer with the given prefix. 

    flat means only generate one level of a directory,
    while non-flat means generate all file paths with that 
    prefix.

    Here's how flat=True handles different senarios:
      1. partial directory name prefix = 'bigarr'
        - lists the '' directory and filters on key 'bigarr'
      2. full directory name prefix = 'bigarray'
        - Same as (1), but using key 'bigarray'
      3. full directory name + "/" prefix = 'bigarray/'
        - Lists the 'bigarray' directory
      4. partial file name prefix = 'bigarray/chunk_'
        - Lists the 'bigarray/' directory and filters on 'chunk_'
    
    Return: generated sequence of file paths relative to layer_path
    """
    with self.get_connection() as conn:
      for f in conn.list_files(prefix, flat):
        yield f

  def __enter__(self):
    StorageBase.__enter__(self)
    self.start_threads()
    return self

  def __exit__(self, exception_type, exception_value, traceback):
    StorageBase.__exit__(self, exception_type, exception_value, traceback)
    self.pool.join()
    self.kill_threads()

class ThreadedStorage(StorageBase, ThreadedQueue):
  """
  Access files stored in Google Storage (gs), Amazon S3 (s3), 
  or the local Filesystem (file).

  e.g. with Storage('gs://bucket/dataset/layer') as stor:
      files = stor.get_file('filename')

  Required:
    layer_path (str): A protocol prefixed path of the above format.
      Accepts s3:// gs:// and file://. File paths are absolute.

  Optional:
    n_threads (int:20): number of threads to use downloading and uplaoding.
      If 0, execution will be on the main python thread.
    progress (bool:false): Show a tqdm progress bar for multiple 
      uploads and downloads.
  """
  def __init__(self, layer_path, n_threads=20, progress=False):
    StorageBase.__init__(self, layer_path, progress)
    ThreadedQueue.__init__(self, n_threads)
    self._interface = self.get_connection()

  def _initialize_interface(self):
    return self._interface_cls(self._path)

  def _close_interface(self, interface):
    interface.release_connection()

  def _consume_queue(self, terminate_evt):
    ThreadedQueue._consume_queue(self, terminate_evt)
    self._interface.release_connection()

  @property
  def layer_path(self):
    return self._layer_path

  def get_path_to_file(self, file_path):
    return posixpath.join(self._layer_path, file_path)

  def put_json(self, file_path, content, content_type='application/json', *args, **kwargs):
    if type(content) != str:
      content = jsonify(content)
    return self.put_file(file_path, content, content_type=content_type, *args, **kwargs)
  
  def put_file(self, file_path, content, content_type=None, compress=None, compress_level=None, cache_control=None):
    """ 
    Args:
      filename (string): it can contains folders
      content (string): binary data to save
    """
    return self.put_files([ (file_path, content) ], 
      content_type=content_type, 
      compress=compress, 
      compress_level=compress_level,
      cache_control=cache_control, 
      block=False
    )

  def put_files(self, files, content_type=None, compress=None, compress_level=None, cache_control=None, block=True):
    """
    Put lots of files at once and get a nice progress bar. It'll also wait
    for the upload to complete, just like get_files.

    Required:
      files: [ (filepath, content), .... ]
    """
    def base_uploadfn(path, content, interface):
      interface.put_file(path, content, content_type, compress, cache_control=cache_control)

    seen = set() 
    for path, content in files:
      if path in seen:
        raise ValueError("Cannot write the same file multiple times in one pass. This causes a race condition. File: " + path)
      seen.add(path)

      content = compression.compress(content, method=compress, compress_level=compress_level)
      uploadfn = partial(base_uploadfn, path, content)

      if len(self._threads):
        self.put(uploadfn)
      else:
        uploadfn(self._interface)

    if block:
      desc = self.progress_description('Uploading')
      self.wait(desc)

    return self

  def exists(self, file_path):
    """Test if a single file exists. Returns boolean."""
    return self._interface.exists(file_path)

  def files_exist(self, file_paths):
    """
    Threaded exists for all file paths.

    file_paths: (list) file paths to test for existence

    Returns: { filepath: bool }
    """
    results = {}

    def exist_thunk(paths, interface):
      results.update(interface.files_exist(paths))

    if len(self._threads):
      for block in scatter(file_paths, len(self._threads)):
        self.put(partial(exist_thunk, block))
    else:
      exist_thunk(file_paths, self._interface)

    desc = self.progress_description('Existence Testing')
    self.wait(desc)

    return results

  def get_file(self, file_path, start=None, end=None):
    """
    Get the binary contents of a file. Optionally, specify
    the inclusive byte range to request.
    """
    content, encoding = self._interface.get_file(file_path, start=start, end=end)
    content = compression.decompress(content, encoding, filename=file_path)
    return content

  def get_files(self, file_paths, starts=None, ends=None):
    """
    returns a list of files faster by using threads
    """
    results = []
    starts, ends = default_byte_iterator(starts, ends)

    def get_file_thunk(path, start, end, interface):
      result = error = None 

      try:
        result = interface.get_file(path, start=start, end=end)
      except Exception as err:
        error = err
        # important to print immediately because 
        # errors are collected at the end
        print(err) 
      
      content, encoding = result
      content = compression.decompress(content, encoding)

      results.append({
        "filename": path,
        "byte_range": (start, end),
        "content": content,
        "error": error,
      })

    for path, start, end in zip(file_paths, starts, ends):
      if len(self._threads):
        self.put(partial(get_file_thunk, path, start, end))
      else:
        get_file_thunk(path, start, end, self._interface)

    desc = self.progress_description('Downloading')
    self.wait(desc)

    return results

  def delete_file(self, file_path):

    def thunk_delete(interface):
      interface.delete_file(file_path)

    if len(self._threads):
      self.put(thunk_delete)
    else:
      thunk_delete(self._interface)

    return self

  def delete_files(self, file_paths):

    def thunk_delete(path, interface):
      interface.delete_file(path)

    for path in file_paths:
      if len(self._threads):
        self.put(partial(thunk_delete, path))
      else:
        thunk_delete(path, self._interface)

    desc = self.progress_description('Deleting')
    self.wait(desc)

    return self

  def list_files(self, prefix="", flat=False):
    """
    List the files in the layer with the given prefix. 

    flat means only generate one level of a directory,
    while non-flat means generate all file paths with that 
    prefix.

    Here's how flat=True handles different senarios:
      1. partial directory name prefix = 'bigarr'
        - lists the '' directory and filters on key 'bigarr'
      2. full directory name prefix = 'bigarray'
        - Same as (1), but using key 'bigarray'
      3. full directory name + "/" prefix = 'bigarray/'
        - Lists the 'bigarray' directory
      4. partial file name prefix = 'bigarray/chunk_'
        - Lists the 'bigarray/' directory and filters on 'chunk_'
    
    Return: generated sequence of file paths relative to layer_path
    """

    for f in self._interface.list_files(prefix, flat):
      yield f

  def __del__(self):
    ThreadedQueue.__del__(self)
    self._interface.release_connection()

  def __exit__(self, exception_type, exception_value, traceback):
    ThreadedQueue.__exit__(self, exception_type, exception_value, traceback)
    self._interface.release_connection()





# @license
# Copyright 2017 The Neuroglancer Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import zlib
import io
import numpy as np
import simplejpeg

from .lib import yellow

try:
  import compressed_segmentation as cseg
  ACCELERATED_CSEG = True # C extension version
except ImportError:
  ACCELERATED_CSEG = False # Pure Python implementation

from . import py_compressed_segmentation as csegpy

try:
  import fpzip 
except ImportError:
  fpziperrormsg = yellow("CloudVolume: fpzip codec is not available. Was it installed? pip install fpzip")
  class fpzip():
    @classmethod
    def compress(cls, content):
      raise NotImplementedError(fpziperrormsg)
    @classmethod
    def decompress(cls, content):
      raise NotImplementedError(fpziperrormsg)

def encode(img_chunk, encoding, block_size=None):
  if encoding == "raw":
    return encode_raw(img_chunk)
  elif encoding == "kempressed":
    return encode_kempressed(img_chunk)
  elif encoding == "fpzip":
    img_chunk = np.asfortranarray(img_chunk)
    return fpzip.compress(img_chunk, order='F')
  elif encoding == "compressed_segmentation":
    return encode_compressed_segmentation(img_chunk, block_size=block_size)
  elif encoding == "jpeg":
    return encode_jpeg(img_chunk)
  elif encoding == "npz":
    return encode_npz(img_chunk)
  elif encoding == "npz_uint8":
    chunk = img_chunk * 255
    chunk = chunk.astype(np.uint8)
    return encode_npz(chunk)
  else:
    raise NotImplementedError(encoding)

def decode(filedata, encoding, shape=None, dtype=None, block_size=None, 
                     background_color=0):
  if (shape is None or dtype is None) and encoding not in ('npz', 'fpzip', 'kempressed'):
    raise ValueError("Only npz encoding can omit shape and dtype arguments. {}".format(encoding))

  if filedata is None or len(filedata) == 0:
    return np.full(shape=shape, fill_value=background_color, dtype=dtype)
  elif encoding == "raw":
    return decode_raw(filedata, shape=shape, dtype=dtype)
  elif encoding == "kempressed":
    return decode_kempressed(filedata)
  elif encoding == "fpzip":
    return fpzip.decompress(filedata, order='F')
  elif encoding == "compressed_segmentation":
    return decode_compressed_segmentation(filedata, shape=shape, dtype=dtype, block_size=block_size)
  elif encoding == "jpeg":
    return decode_jpeg(filedata, shape=shape, dtype=dtype)
  elif encoding == "npz":
    return decode_npz(filedata)
  else:
    raise NotImplementedError(encoding)

def encode_jpeg(arr, quality=85):
  if not np.issubdtype(arr.dtype, np.uint8):
    raise ValueError("Only accepts uint8 arrays. Got: " + str(arr.dtype))

  # simulate multi-channel array for single channel arrays
  while arr.ndim < 4:
    arr = arr[..., np.newaxis] # add channels to end of x,y,z

  num_channel = arr.shape[3]
  reshaped = arr.T
  reshaped = np.moveaxis(reshaped, 0, -1)
  reshaped = reshaped.reshape(
    arr.shape[0] * arr.shape[1], arr.shape[2], num_channel
  )
  if num_channel == 1:
    return simplejpeg.encode_jpeg(
      np.ascontiguousarray(reshaped), 
      colorspace="GRAY",
      colorsubsampling="GRAY",
      quality=quality,
    )
  elif num_channel == 3:
    return simplejpeg.encode_jpeg(
      np.ascontiguousarray(reshaped),
      colorspace="RGB",
      quality=quality,
    )
  raise ValueError("Number of image channels should be 1 or 3. Got: {}".format(arr.shape[3]))

def encode_npz(subvol):
  """
  This file format is unrelated to np.savez
  We are just saving as .npy and the compressing
  using zlib. 
  The .npy format contains metadata indicating
  shape and dtype, instead of np.tobytes which doesn't
  contain any metadata.
  """
  fileobj = io.BytesIO()
  if len(subvol.shape) == 3:
    subvol = np.expand_dims(subvol, 0)
  np.save(fileobj, subvol)
  cdz = zlib.compress(fileobj.getvalue())
  return cdz

def encode_compressed_segmentation(subvol, block_size, accelerated=ACCELERATED_CSEG):
  assert np.dtype(subvol.dtype) in (np.uint32, np.uint64)

  if accelerated:
    return encode_compressed_segmentation_c_ext(subvol, block_size)  
  return encode_compressed_segmentation_pure_python(subvol, block_size)

def encode_compressed_segmentation_c_ext(subvol, block_size):
  if np.dtype(subvol.dtype) not in (np.uint32, np.uint64):
    raise ValueError("compressed_segmentation only supports uint32 and uint64 datatypes. Got: " + str(subvol.dtype))

  subvol = np.squeeze(subvol, axis=3)
  if subvol.flags.c_contiguous:
    order = 'C' 
  elif subvol.flags.f_contiguous:
    order = 'F'
  else:
    order = 'F'
    subvol = np.asfortranarray(subvol)
  return cseg.compress(subvol, block_size=block_size, order=order)

def encode_compressed_segmentation_pure_python(subvol, block_size):
  return csegpy.encode_chunk(subvol.T, block_size=block_size)

def encode_raw(subvol):
  return subvol.tobytes('F')

def encode_kempressed(subvol):
  data = 2.0 + np.swapaxes(subvol, 2,3)
  return fpzip.compress(data, order='F')

def decode_kempressed(bytestring):
  """subvol not bytestring since numpy conversion is done inside fpzip extension."""
  subvol = fpzip.decompress(bytestring, order='F')
  return np.swapaxes(subvol, 3,2) - 2.0

def decode_npz(string):
  fileobj = io.BytesIO(zlib.decompress(string))
  return np.load(fileobj)

def decode_jpeg(bytestring, shape, dtype):
  colorspace = "RGB" if len(shape) > 3 and shape[3] > 1 else "GRAY"
  data = simplejpeg.decode_jpeg(
    bytestring, 
    colorspace=colorspace,
  ).ravel()
  return data.reshape(shape, order='F')

def decode_raw(bytestring, shape, dtype):
  return np.frombuffer(bytearray(bytestring), dtype=dtype).reshape(shape, order='F')

def decode_compressed_segmentation(bytestring, shape, dtype, block_size, accelerated=ACCELERATED_CSEG):
  if block_size is None:
    raise ValueError("block_size parameter must not be None.")

  if accelerated:
    return decode_compressed_segmentation_c_ext(bytestring, shape, dtype, block_size)

  return decode_compressed_segmentation_pure_python(bytestring, shape, dtype, block_size)

def decode_compressed_segmentation_c_ext(bytestring, shape, dtype, block_size):
  return cseg.decompress(bytes(bytestring), shape, dtype, block_size, order='F')

def decode_compressed_segmentation_pure_python(bytestring, shape, dtype, block_size):
  chunk = np.empty(shape=shape[::-1], dtype=dtype)
  csegpy.decode_chunk_into(chunk, bytestring, block_size=block_size)
  return chunk.T




import python_jsonschema_objects as pjs
import json
import json5

__all__ = [ 'DatasetProvenance', 'DataLayerProvenance' ]

dataset_provenance_schema = {
  "$schema": "http://json-schema.org/draft-04/schema#",
  "title": "Dataset Provenance",
  "description": "Represents a dataset and its derived data layers.",
  "required": [
    "dataset_name", "dataset_description", 
    "organism", "imaged_date", "imaged_by", 
    "owners"
  ],
  "properties": {
    'dataset_name': { 'type': 'string' },
    'dataset_description': { 'type': 'string' },
    'organism': {
      'type': 'string',
      'description': 'Species, sex, strain identifier',
    },
    'imaged_date': { 'type': 'string' },
    'imaged_by': { 'type': 'string' },
    'references': { # e.g. dois, urls, titles
      "type": "array",
      "items": {
        "type": "string"
      },
      "minItems": 0,
      "uniqueItems": True, # e.g. email addresses  
    }, 
    'owners': {
      "type": "array",
      "items": {
        "type": "string"
      },
      "minItems": 1,
      "uniqueItems": True, # e.g. email addresses  
    }
  }
}

builder = pjs.ObjectBuilder(dataset_provenance_schema)
classes = builder.build_classes()
DatasetProvenance = classes.DatasetProvenance

layer_provenance_schema = {
  "$schema": "http://json-schema.org/draft-04/schema#",
  "title": "Data Layer Provenance",
  "description": "Represents a data layer within a dataset. e.g. image, segmentation, etc",
  "required": [
    'description', 'sources', 
    'processing', 'owners'
  ],
  "properties": {
    'description': { 'type': 'string' },
    'sources': { # e.g. [ 'gs://neuroglancer/pinky40_v11/image' 
      "type": "array",
      "items": {
        "type": "string"
      },
      "minItems": 0,
      "uniqueItems": True,
    }, 
    'processing': {
      "type": "array",
      "items": {
        "type": "object"
      },
      "minItems": 0,
    },
    # e.g. processing = [ 
    #    { 'method': 'inceptionnet', 'by': 'example@princeton.edu' }, 
    #    { 'method': 'downsample', 'by': 'example2@princeton.edu', 'description': 'demo of countless downsampling' } 
    #    { 'method': 'meshing', 'by': 'example2@princeton.edu', 'description': '512x512x512 mip 3 simplification factor 30' }
    # ]
    'owners': {
      "type": "array",
      "items": {
        "type": "string"
      },
      "minItems": 0,
      "uniqueItems": True,
    },
  }
}

builder = pjs.ObjectBuilder(layer_provenance_schema)
classes = builder.build_classes()
DataLayerProvenanceValidation = classes.DataLayerProvenance

class DataLayerProvenance(dict):
  def __init__(self, *args, **kwargs):
    dict.__init__(self, *args, **kwargs)
    if 'description' not in self:
      self['description'] = ''
    if 'owners' not in self:
      self['owners'] = []
    if 'processing' not in self:
      self['processing'] = []
    if 'sources' not in self:
      self['sources'] = ''

  def validate(self):
    DataLayerProvenanceValidation(**self).validate()

  def serialize(self):
    return json.dumps(self)

  def from_json(self, data):
    data = json5.loads(data)
    self.update(data)
    return self

  @classmethod
  def create(cls, mydict):
    return DatasetProvenance(**mydict)

  @property 
  def description(self):
    return self['description']
  @description.setter 
  def description(self, val):
    self['description'] = val
  
  @property 
  def owners(self):
    return self['owners']
  @owners.setter 
  def owners(self, val):
    self['owners'] = val

  @property 
  def processing(self):
    return self['processing']
  @processing.setter 
  def processing(self, val):
    self['processing'] = val

  @property 
  def sources(self):
    return self['sources']
  @sources.setter 
  def sources(self, val):
    self['sources'] = val
    













import copy
import re
import struct

import numpy as np

from .exceptions import MeshDecodeError
from .lib import yellow, Vec

NOTICE = {
  'vertices': 0,
  'num_vertices': 0,
  'faces': 0,
}

def deprecation_notice(key):
  if NOTICE[key] < 1:
    print(yellow("""
  Deprecation Notice: Meshes, formerly dicts, are now PrecomputedMesh objects
  as of CloudVolume 0.51.0, renamed to Mesh objects as of 0.53.0

  Please change mesh['{}'] to mesh.{}
  """.format(key, key)))
    NOTICE[key] += 1

def is_draco_chunk_aligned(verts, chunk_size, draco_grid_size):
  """
  Return a mask that for each vertex is true iff it is within
  half a draco_grid_size from a chunk border.
  """
  dist_to_chunk_behind = np.mod(verts, chunk_size)
  dist_to_chunk_ahead = chunk_size - dist_to_chunk_behind
  # Draco rounds up
  is_on_chunk_behind = np.any(
    dist_to_chunk_behind < (draco_grid_size / 2),
    axis=1,
  )
  is_on_chunk_ahead = np.any(
    dist_to_chunk_ahead <= (draco_grid_size / 2),
    axis=1,
  )
  return np.logical_or(is_on_chunk_behind, is_on_chunk_ahead)
    

class Mesh(object):
  """
  Represents the vertices, faces, and normals of a mesh
  as numpy arrays.

  class Mesh:
    ndarray[float32, ndim=2] self.vertices: [ [x,y,z], ... ]
    ndarray[uint32,  ndim=2] self.faces:    [ [v1,v2,v3], ... ]
    ndarray[float32, ndim=2] self.normals:  [ [nx,ny,nz], ... ]
  """
  def __init__(
    self, vertices, faces, normals=None, 
    segid=None, encoding_type=None, encoding_options=None
  ):
    self.vertices = np.array(vertices, dtype=np.float32)
    self.faces = np.array(faces, dtype=np.uint32)

    if normals is None:
      self.normals = np.array([], dtype=np.float32).reshape((0,3))
    else:
      self.normals = np.array(normals, dtype=np.float32)

    self.segid = segid
    self.encoding_type = encoding_type
    self.encoding_options = encoding_options 

  def __len__(self):
    return self.vertices.shape[0]

  def __eq__(self, other):
    """Tests strict equality between two meshes."""

    no_self_normals = self.normals is None or self.normals.size == 0
    no_other_normals = other.normals is None or other.normals.size == 0

    if no_self_normals != no_other_normals:
      return False
       
    equality = np.all(self.vertices == other.vertices) \
      and np.all(self.faces == other.faces)

    if no_self_normals:
      return equality

    return (equality and np.all(self.normals == other.normals))

  def __repr__(self):
    return "Mesh(vertices<{}>, faces<{}>, normals<{}>, segid={}, encoding_type=<{}>)".format(
      self.vertices.shape[0], self.faces.shape[0], self.normals.shape[0],
      self.segid, self.encoding_type
    )

  def __getitem__(self, key):
    val = None 
    if key == 'vertices':
      val = self.vertices
    elif key == 'num_vertices':
      val = len(self)
    elif key == 'faces':
      val = self.faces
    else:
      raise KeyError("{} not found.".format(key))

    deprecation_notice(key)
    return val

  def empty(self):
    return self.vertices.size == 0 or self.faces.size == 0

  def clone(self):
    return Mesh(
      np.copy(self.vertices), np.copy(self.faces), np.copy(self.normals),
      self.segid, 
      encoding_type=copy.deepcopy(self.encoding_type),
      encoding_options=copy.deepcopy(self.encoding_options),
    )

  def triangles(self):
    Nf = self.faces.shape[0]
    tris = np.zeros( (Nf, 3, 3), dtype=np.float32, order='C' ) # triangle, vertices, (x,y,z)

    for i in range(Nf):
      for j in range(3):
        tris[i,j,:] = self.vertices[ self.faces[i,j] ]

    return tris

  @classmethod
  def concatenate(cls, *meshes):
    vertex_ct = np.zeros(len(meshes) + 1, np.uint32)
    vertex_ct[1:] = np.cumsum([ len(mesh) for mesh in meshes ])

    vertices = np.concatenate([ mesh.vertices for mesh in meshes ])
    
    faces = np.concatenate([ 
      mesh.faces + vertex_ct[i] for i, mesh in enumerate(meshes) 
    ])

    normals = np.concatenate([ mesh.normals for mesh in meshes ])

    encoding_type = list(set([ mesh.encoding_type for mesh in meshes ]))
    if len(encoding_type) == 1:
      encoding_type = encoding_type[0]

    return Mesh(vertices, faces, normals, encoding_type=encoding_type)

  def consolidate(self):
    """Remove duplicate vertices and faces. Returns a new mesh object."""
    if self.empty():
      return Mesh([], [], normals=None)

    vertices = self.vertices
    faces = self.faces
    normals = self.normals

    eff_verts, uniq_idx, idx_representative = np.unique(
      vertices, axis=0, return_index=True, return_inverse=True
    )

    face_vector_map = np.vectorize(lambda x: idx_representative[x])
    eff_faces = face_vector_map(faces)
    eff_faces = np.unique(eff_faces, axis=0)

    # normal_vector_map = np.vectorize(lambda idx: normals[idx])
    # eff_normals = normal_vector_map(uniq_idx)

    return Mesh(eff_verts, eff_faces, None, 
      segid=self.segid,
      encoding_type=copy.deepcopy(self.encoding_type),
      encoding_options=copy.deepcopy(self.encoding_options),
    )

  @classmethod
  def from_precomputed(self, binary, segid=None):
    """
    Mesh from_precomputed(self, binary)

    Decode a Precomputed format mesh from a byte string.
    
    Format:
      uint32        Nv * float32 * 3   uint32 * 3 until end
      Nv            (x,y,z)            (v1,v2,v2)
      N Vertices    Vertices           Faces
    """
    num_vertices = struct.unpack("=I", binary[0:4])[0]
    try:
      # count=-1 means all data in buffer
      vertices = np.frombuffer(binary, dtype=np.float32, count=3*num_vertices, offset=4)
      faces = np.frombuffer(binary, dtype=np.uint32, count=-1, offset=(4 + 12 * num_vertices)) 
    except ValueError:
      raise MeshDecodeError("""
        The input buffer is too small for the Precomputed format.
        Minimum Bytes: {} 
        Actual Bytes: {}
      """.format(4 + 4 * num_vertices, len(binary)))

    vertices = vertices.reshape(num_vertices, 3)
    faces = faces.reshape(faces.size // 3, 3)

    return Mesh(
      vertices, faces, 
      segid=segid, 
      normals=None, 
      encoding_type='precomputed'
    )

  def to_precomputed(self):
    """
    bytes to_precomputed(self)

    Convert mesh into binary format compatible with Neuroglancer.
    Does not preserve normals.
    """
    vertex_index_format = [
      np.uint32(self.vertices.shape[0]), # Number of vertices (3 coordinates)
      self.vertices,
      self.faces
    ]
    return b''.join([ array.tobytes('C') for array in vertex_index_format ])

  @classmethod
  def from_obj(self, text, segid=None):
    """Given a string representing a Wavefront OBJ file, decode to a Mesh."""

    vertices = []
    faces = []
    normals = []

    if type(text) is bytes:
      text = text.decode('utf8')

    for line in text.split('\n'):
      line = line.strip()
      if len(line) == 0:
        continue
      elif line[0] == '#':
        continue
      elif line[0] == 'f':
        if line.find('/') != -1:
          # e.g. f 6092/2095/6079 6087/2092/6075 6088/2097/6081
          (v1, vt1, vn1, v2, vt2, vn2, v3, vt3, vn3) = re.match(r'f\s+(\d+)/(\d*)/(\d+)\s+(\d+)/(\d*)/(\d+)\s+(\d+)/(\d*)/(\d+)', line).groups()
        else:
          (v1, v2, v3) = re.match(r'f\s+(\d+)\s+(\d+)\s+(\d+)', line).groups()
        faces.append( (int(v1), int(v2), int(v3)) )
      elif line[0] == 'v':
        if line[1] == 't': # vertex textures not supported
          # e.g. vt 0.351192 0.337058
          continue 
        elif line[1] == 'n': # vertex normals
          # e.g. vn 0.992266 -0.033290 -0.119585
          (n1, n2, n3) = re.match(r'vn\s+([-\d\.]+)\s+([-\d\.]+)\s+([-\d\.]+)', line).groups()
          normals.append( (float(n1), float(n2), float(n3)) )
        else:
          # e.g. v -0.317868 -0.000526 -0.251834
          (v1, v2, v3) = re.match(r'v\s+([-\d\.]+)\s+([-\d\.]+)\s+([-\d\.]+)', line).groups()
          vertices.append( (float(v1), float(v2), float(v3)) )

    vertices = np.array(vertices, dtype=np.float32)
    faces = np.array(faces, dtype=np.uint32)
    normals = np.array(normals, dtype=np.float32)

    return Mesh(vertices, faces - 1, normals, segid=segid)

  def to_obj(self):
    """Return a string representing a .obj file."""
    objdata = []
    objdata += [ 'v {:.5f} {:.5f} {:.5f}'.format(*vertex) for vertex in self.vertices ]
    objdata += [ 'f {} {} {}'.format(*face) for face in (self.faces+1) ] # obj is 1 indexed
    objdata = '\n'.join(objdata) + '\n'
    return objdata.encode('utf8')

  def to_ply(self):
    """
    Return a bytearray in .ply format, 
    a more compact format than .obj that's still widely readable.
    """
    vertexct = self.vertices.shape[0]
    trianglect = self.faces.shape[0]

    # Header
    plydata = bytearray("""ply
format binary_little_endian 1.0
element vertex {}
property float x
property float y
property float z
element face {}
property list int int vertex_indices
end_header
""".format(vertexct, trianglect).encode('utf8'))

    # Vertex data (x y z): "fff" 
    plydata.extend(self.vertices.tobytes('C'))

    # Faces (3 f1 f2 f3): "3iii" 
    plydata.extend(
      np.insert(self.faces, 0, 3, axis=1).tobytes('C')
    )

    return plydata

  @classmethod
  def from_draco(cls, binary, segid=None):
    import DracoPy

    try:
      mesh_object = DracoPy.decode_buffer_to_mesh(binary)
      vertices = np.array(mesh_object.points).astype(np.float32)
      faces = np.array(mesh_object.faces).astype(np.uint32)
    except ValueError:
      raise MeshDecodeError("Not a valid draco mesh.")

    Nv = len(vertices)
    Nf = len(faces)

    if Nv % 3 != 0: 
      raise MeshDecodeError("Mesh vertices not 3D. Cannot decode. Num. vertices: {}".format(Nv))

    if Nf % 3 != 0:
      raise MeshDecodeError("Faces are not sets of triples. Cannot decode. Num. faces: {}".format(Nf))

    vertices = vertices.reshape(Nv // 3, 3)
    faces = faces.reshape(Nf // 3, 3)

    return Mesh(
      vertices, faces, 
      segid=segid,
      normals=None,
      encoding_type='draco', 
      encoding_options=mesh_object.encoding_options
    )

  def deduplicate_vertices(self, is_chunk_aligned):
    faces = self.faces
    verts = self.vertices
    # find all vertices that have exactly 2 duplicates
    unique_vertices, unique_inverse, counts = np.unique(
      verts, return_inverse=True, return_counts=True, axis=0
    )
    
    only_double = np.where(counts == 2)[0]
    is_doubled = np.isin(unique_inverse, only_double)
    # this stores whether each vertex should be merged or not
    do_merge = np.array(is_doubled & is_chunk_aligned)

    # setup an artificial 4th coordinate for vertex positions
    # which will be unique in general, 
    # but then repeated for those that are merged
    new_vertices = np.hstack((verts, np.arange(verts.shape[0])[:, np.newaxis]))
    new_vertices[do_merge, 3] = -1
  
    faces = faces.flatten()

    # use unique to make the artificial vertex list unique and reindex faces
    vertices, newfaces = np.unique(new_vertices[faces], return_inverse=True, axis=0)
    newfaces = newfaces.astype(np.uint32).reshape( (len(newfaces) // 3, 3) )

    return Mesh(vertices[:,0:3], newfaces, None, segid=self.segid, 
      encoding_type=self.encoding_type, encoding_options=self.encoding_options
    )

  def deduplicate_chunk_boundaries(self, chunk_size, is_draco=False, draco_grid_size=None, offset=(0,0,0)):
    offset = Vec(*offset)
    verts = self.vertices - offset
    # find all vertices that are exactly on chunk_size boundaries
    if is_draco:
      if draco_grid_size is None:
        raise ValueError('Must specify draco grid size to dedup draco meshes')
      is_chunk_aligned = is_draco_chunk_aligned(verts, chunk_size, draco_grid_size=draco_grid_size)
    else:
      is_chunk_aligned = np.any(np.mod(verts, chunk_size) == 0, axis=1)

    return self.deduplicate_vertices(is_chunk_aligned)

  def viewer(self):
    try:
      import vtk
    except ImportError:
      print("The mesh viewer requires the OpenGL based vtk. Try: pip install vtk --upgrade")
      return

    colors = vtk.vtkNamedColors()
    # Set the background color.
    bkg = map(lambda x: x / 255.0, [40, 40, 40, 255])
    colors.SetColor("BkgColor", *bkg)

    # This creates a point cloud model
    points = vtk.vtkPoints()
    verts = vtk.vtkCellArray()
    polydata = vtk.vtkPolyData()
    polydata.SetPoints(points)
    polydata.SetVerts(verts)
    for vertex in self.vertices:
      pid = points.InsertNextPoint(vertex)
      verts.InsertNextCell(1)
      verts.InsertCellPoint(pid)

    # The mapper is responsible for pushing the geometry into the graphics
    # library. It may also do color mapping, if scalars or other
    # attributes are defined.
    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputData(polydata)

    # The actor is a grouping mechanism: besides the geometry (mapper), it
    # also has a property, transformation matrix, and/or texture map.
    # Here we set its color and rotate it -22.5 degrees.
    cylinderActor = vtk.vtkActor()
    cylinderActor.SetMapper(mapper)
    cylinderActor.GetProperty().SetColor(colors.GetColor3d("Mint"))
    cylinderActor.RotateX(30.0)
    cylinderActor.RotateY(-45.0)

    # Create the graphics structure. The renderer renders into the render
    # window. The render window interactor captures mouse events and will
    # perform appropriate camera or actor manipulation depending on the
    # nature of the events.
    ren = vtk.vtkRenderer()
    renWin = vtk.vtkRenderWindow()
    renWin.AddRenderer(ren)
    iren = vtk.vtkRenderWindowInteractor()
    iren.SetRenderWindow(renWin)

    # Add the actors to the renderer, set the background and size
    ren.AddActor(cylinderActor)
    ren.SetBackground(colors.GetColor3d("BkgColor"))
    renWin.SetSize(800, 800)

    text = "Point Cloud Rendering of Mesh Object"
    if self.segid is not None:
      renWin.SetWindowName(text + " (Label {})".format(self.segid))
    else:
      renWin.SetWindowName(text)

    # This allows the interactor to initalize itself. It has to be
    # called before an event loop.
    iren.Initialize()

    # We'll zoom in a little by accessing the camera and invoking a "Zoom"
    # method on it.
    ren.ResetCamera()
    ren.GetActiveCamera().Zoom(1.5)
    renWin.Render()

    # Start the event loop.
    iren.Start()    



from collections import defaultdict
import simdjson
import os 
import sqlite3

import numpy as np
from tqdm import tqdm

from cloudfiles import CloudFiles

from ...exceptions import SpatialIndexGapError
from ... import paths
from ...lib import Bbox, Vec, xyzrange, min2, toiter, sip, nvl

class SpatialIndex(object):
  """
  Implements the client side reader of the 
  spatial index. During data generation, the
  labels in a given task are enumerated and 
  assigned their bounding box as JSON:

  {
    SEGID: [ x,y,z, x,y,z ],
    ...
  }

  The filename is the physical bounding box of the
  task dot spatial.

  e.g. "0-1024_0-1024_0-500.spatial" where the bbox 
  units are nanometers.

  The info file of the data type can then be augmented
  with:

  {
    "spatial_index": { "chunk_size": [ sx, sy, sz ] }
  }

  Where sx, sy, and sz are given in physical dimensions.
  """
  def __init__(
    self, cloudpath, bounds, chunk_size, 
    config=None, sqlite_db=None
  ):
    self.cloudpath = cloudpath
    self.path = paths.extract(cloudpath)
    self.bounds = Bbox.create(bounds)
    self.chunk_size = Vec(*chunk_size)
    self.sqlite_db = sqlite_db # optional DB for higher performance

    if config is None:
      self.config = {}
      self.config.progress = None
    else:
      self.config = config

  def join(self, *paths):
    if self.path.protocol == 'file':
      return os.path.join(*paths)
    else:
      return posixpath.join(*paths)    

  def fetch_index_files(self, index_files, progress=None):
    progress = nvl(progress, self.config.progress)
    results = CloudFiles(self.cloudpath, progress=progress).get(index_files)

    for res in results:
      if res['error'] is not None:
        raise SpatialIndexGapError(res['error'])

    return { res['filename']: res['content'] for res in results }

  def fetch_all_index_files(self, allow_missing=False, progress=None):
    """Generator returning batches of (filename, json)"""
    all_index_paths = self.index_file_paths_for_bbox(self.bounds)
    
    progress = nvl(progress, self.config.progress)

    N = 500
    pbar = tqdm( 
      total=len(all_index_paths), 
      disable=(not progress), 
      desc="Processing Index"
    )

    for index_paths in sip(all_index_paths, N):
      index_files = self.fetch_index_files(index_paths, progress=False)

      for filename, content in index_files.items():
        if content is None:
          if allow_missing:
            continue
          else:
            raise SpatialIndexGapError(filename + " was not found.")

      yield index_files

      pbar.update(N)
    pbar.close()

  def index_file_paths_for_bbox(self, bbox):
    bbox = bbox.expand_to_chunk_size(self.chunk_size, offset=self.bounds.minpt)

    if bbox.subvoxel():
      return []

    chunk_size = self.chunk_size
    bounds = self.bounds

    class IndexPathIterator():
      def __len__(self):
        return bbox.num_chunks(chunk_size)
      def __iter__(self):
        for pt in xyzrange(bbox.minpt, bbox.maxpt, chunk_size):
          search = Bbox( pt, min2(pt + chunk_size, bounds.maxpt) )
          yield search.to_filename() + '.spatial'

    return IndexPathIterator()

  def to_sqlite(
    self, database_name="spatial_index.db", 
    create_indices=True, progress=None
  ):
    """
    Create a sqlite database of labels and filenames
    from the JSON spatial_index for faster performance.

    Depending on the dataset size, this could take a while.
    With a dataset with ~140k index files, the DB took over
    an hour to build and was 42 GB.
    """
    progress = nvl(progress, self.config.progress)

    conn = sqlite3.connect(database_name)
    cur = conn.cursor()

    cur.execute("""
    CREATE TABLE index_files (
      id INTEGER PRIMARY KEY AUTOINCREMENT,
      filename TEXT NOT NULL
    )
    """)
    cur.execute("CREATE INDEX idxfname ON index_files (filename)")

    cur.execute("""
    CREATE TABLE file_lookup (
      label INTEGER NOT NULL,
      fid INTEGER NOT NULL REFERENCES index_files(id),
      PRIMARY KEY(label,fid)
    )
    """)

    cur.execute("PRAGMA journal_mode = MEMORY")
    cur.execute("PRAGMA synchronous = OFF")

    parser = simdjson.Parser()

    for index_files in self.fetch_all_index_files(progress=progress):
      for filename, content in index_files.items():
        index_labels = parser.parse(content).keys()
        filename = os.path.basename(filename)
        cur.execute("INSERT INTO index_files(filename) VALUES (?)", (filename,))
        cur.execute("SELECT id from index_files where filename = ?", (filename,))
        fid = cur.fetchone()[0]
        values = ( (int(label), fid) for label in index_labels )
        cur.executemany("INSERT INTO file_lookup(label, fid) VALUES (?,?)", values)
      conn.commit()

    cur.execute("PRAGMA journal_mode = DELETE")
    cur.execute("PRAGMA synchronous = FULL")

    if create_indices:
      if progress:
        print("Creating labels index...")
      cur.execute("CREATE INDEX file_lbl ON file_lookup (label)")

      if progress:
        print("Creating filename index...")
      cur.execute("CREATE INDEX fname ON file_lookup (fid)")

    conn.close()

  def get_bbox(self, label):
    """
    Given a label, compute an enclosing bounding box for it.

    Returns: Bbox in physical coordinates
    """
    locations = defaultdict(list)
    parser = simdjson.Parser()

    label = str(label)
    bbox = None

    if self.sqlite_db:
      conn = sqlite3.connect(self.sqlite_db)
      cur = conn.cursor()
      cur.execute("""
        select index_files.filename  
        from file_lookup, index_files
        where file_lookup.fid = index_files.id
          and file_lookup.label = ?
      """, (label,))
      iterator = [ self.fetch_index_files(( row[0] for row in cur.fetchall() )) ]
      conn.close()
    else:
      iterator = self.fetch_all_index_files()

    for index_files in iterator:
      for filename, content in index_files.items():
        segid_bbox_dict = parser.parse(content)
        filename = os.path.basename(filename)

        if label not in segid_bbox_dict: 
          continue 

        current_bbox = Bbox.from_list(
          np.frombuffer(segid_bbox_dict[label].as_buffer(of_type="i"), dtype=np.int64)
        )

        if bbox is None:
          bbox = current_bbox
        else:
          bbox = Bbox.expand(bbox, current_bbox)

    return bbox

  def file_locations_per_label(self, labels=None, allow_missing=False):
    """
    Queries entire dataset to find which spatial index files the 
    given labels are located in. Can be expensive. If labels is not 
    specified, all labels are fetched.

    If the spatial_index.sqlite_db attribute is specified, attempt
    to use the database instead of querying the json files.

    Returns: { filename: [ labels... ], ... }
    """
    if labels is not None:
      labels = toiter(labels)
    
    if self.sqlite_db:
      return self.file_locations_per_label_sql(labels)
    return self.file_locations_per_label_json(labels, allow_missing)
  
  def file_locations_per_label_json(self, labels, allow_missing=False):
    locations = defaultdict(list)
    parser = simdjson.Parser()
    if labels is not None:
      labels = set(toiter(labels))

    for index_files in self.fetch_all_index_files():
      for filename, content in index_files.items():
        index_labels = set(parser.parse(content).keys())
        filename = os.path.basename(filename)

        if labels is None:
          for label in index_labels:
            locations[int(label)].append(filename)
        elif len(labels) > len(index_labels):
          for label in index_labels:
            if int(label) in labels:
              locations[int(label)].append(filename)
        else:
          for label in labels:
            if str(label) in index_labels:
              locations[int(label)].append(filename)

    return locations

  def file_locations_per_label_sql(self, labels, sqlite_db=None):
    sqlite_db = nvl(sqlite_db, self.sqlite_db)
    if sqlite_db is None:
      raise ValueError("An sqlite database file must be specified.")

    locations = defaultdict(list)
    conn = sqlite3.connect(sqlite_db)
    cur = conn.cursor()

    where_clause = ""
    if labels:
      where_clause = "and file_lookup.label in ({})".format(
        ",".join(( str(int(lbl)) for lbl in labels ))
      )

    cur.execute("""
      select file_lookup.label, index_files.filename  
      from file_lookup, index_files
      where file_lookup.fid = index_files.id
        {}
    """.format(where_clause))
    while True:
      rows = cur.fetchmany(2**20)
      if len(rows) == 0:
        break
      for label, filename in rows:
        locations[int(label)].append(filename)
    conn.close()
    return locations      

  def query(self, bbox, allow_missing=False):
    """
    For the specified bounding box (or equivalent representation),
    list all segment ids enclosed within it.

    If allow_missing is set, then don't raise an error if an index
    file is missing.

    Returns: set(labels)
    """
    bbox = Bbox.create(bbox, context=self.bounds, autocrop=True)
    original_bbox = bbox.clone()
    bbox = bbox.expand_to_chunk_size(self.chunk_size, offset=self.bounds.minpt)

    if bbox.subvoxel():
      return []

    labels = set()
    fast_path = bbox.contains_bbox(self.bounds)

    if self.sqlite_db and fast_path:
      conn = sqlite3.connect(self.sqlite_db)
      cur = conn.cursor()
      cur.execute("select label from file_lookup")
      while True:
        rows = cur.fetchmany(size=2**20)
        if len(rows) == 0:
          break
        labels.update(( int(row[0]) for row in rows ))
      conn.close()
      return labels

    index_files = self.index_file_paths_for_bbox(bbox)
    results = self.fetch_index_files(index_files)

    parser = simdjson.Parser()
    for filename, content in tqdm(results.items(), desc="Decoding Labels", disable=(not self.config.progress)):
      if content is None:
        if allow_missing:
          continue
        else:
          raise SpatialIndexGapError(filename + " was not found.")

      # The bbox test saps performance a lot
      # but we can skip it if we know 100% that
      # the labels are going to be inside. This
      # optimization is important for querying 
      # entire datasets, which is contemplated
      # for shard generation.
      if fast_path:
        res = parser.parse(content).keys()
        labels.update( (int(label) for label in res ) ) # fast path: 16% CPU
      else:
        res = simdjson.loads(content)
        for label, label_bbx in res.items():
          label = int(label)
          label_bbx = Bbox.from_list(label_bbx)

          if Bbox.intersects(label_bbx, original_bbox):
            labels.add(label)

    return labels

class CachedSpatialIndex(SpatialIndex):
  def __init__(self, cache, config, cloudpath, bounds, chunk_size):
    self.cache = cache
    self.subdir = os.path.relpath(cloudpath, cache.meta.cloudpath)

    super(CachedSpatialIndex, self).__init__(
      cloudpath, bounds, chunk_size, config
    )

  def fetch_index_files(self, index_files, progress=None):
    progress = nvl(progress, self.config.progress)
    index_files = [ self.cache.meta.join(self.subdir, fname) for fname in index_files ]
    return self.cache.download(index_files, progress=progress)



"""
A "serverless" Python client for reading and writing arbitrarily large 
Neuroglancer Precomputed volumes both locally and on cloud services. 
Precomputed volumes consist of chunked numpy arrays, meshes, and 
skeletons and can be visualized using Neuroglancer. Typically these 
volumes represent one or three channel 3D image stacks of microscopy 
data or labels annotating them, but they can also represent large 
tensors with compatible dimensions. 

  https://github.com/seung-lab/cloud-volume

Precomputed volumes can be stored on any service that provides a 
key-value mapping between a file path and file data. Typically, 
Precomputed volumes are located on cloud storage providers such 
as Amazon S3 or Google Cloud Storage. However, these volumes can 
be stored on any service, including the local file system or an 
ordinary webserver that can process these key value mappings.

Neuroglancer is a browser based WebGL 3D image viewer principally 
authored by Jeremy Maitin-Shepard at Google. CloudVolume is a 
third-party client for reading and writing Neuroglancer compatible 
formats: 

  https://github.com/google/neuroglancer

CloudVolume is often paired with Igneous, an image processing engine 
for visualizing and managing Precomputed volumes. Igneous can be
run locally or in the cloud using Kubernetes.

  https://github.com/seung-lab/igneous

The combination of Neuroglancer, Igneous, and CloudVolume comprises 
a system for visualizing, processing, and sharing (via browser viewable 
URLs) petascale datasets within and between laboratories.

CloudVolume Example: 

  from cloudvolume import CloudVolume

  vol = CloudVolume('gs://mylab/mouse/image', progress=True)
  image = vol[:,:,:] # Download an image stack as a numpy array
  vol[:,:,:] = image # Upload an image stack from a numpy array

  label = 1
  mesh = vol.mesh.get(label) 
  skel = vol.skeletons.get(label)
"""

from .cloudvolume import CloudVolume, register_plugin

from .connectionpools import ConnectionPool
from .lib import Bbox, Vec
from .mesh import Mesh
from .provenance import DataLayerProvenance
from .storage import Storage
from .threaded_queue import ThreadedQueue
from .exceptions import (
  EmptyVolumeException, EmptyRequestException, AlignmentError,
  SkeletonEncodeError, SkeletonDecodeError
)
from .volumecutout import VolumeCutout

from .skeleton import Skeleton, PrecomputedSkeleton

from . import exceptions
from . import secrets

from . import microviewer
from .microviewer import view, hyperview

__version__ = '3.8.0'

# Register plugins
from .datasource.precomputed import register as register_precomputed
from .datasource.graphene import register as register_graphene

register_precomputed()
register_graphene()

try:
  from .datasource.boss import register as register_boss
  register_boss()
except ImportError:
  pass









import posixpath
from collections import namedtuple
import json
import re
import requests
from six.moves import urllib

import numpy as np

from ...lib import Vec, Bbox
from ... import exceptions
from ... import paths
from ...secrets import chunkedgraph_credentials
from ..precomputed import PrecomputedMetadata

VERSION_ORDERING = [  
  '1.0', 'v1'
]
VERSION_MAP = {
  version: i for i, version in enumerate(VERSION_ORDERING)
}

uint64 = np.uint64
GrapheneLabel = namedtuple('GrapheneLabel', ('level', 'x', 'y', 'z', 'segid'))


class GrapheneApiVersion():
  def __init__(self, version):
    self.version = version.lower()
    if self.version == 'table':
      self.version = VERSION_ORDERING[-1]
    elif self.version not in VERSION_MAP:
      raise ValueError("Unknown Graphene API version {}".format(self.version))

  def __eq__(self, rhs):
    return self.version == rhs.version
  def __ne__(self, rhs):
    return self.version != rhs.version
  def __lt__(self, rhs):
    return self.sequence_number() < rhs.sequence_number()
  def __gt__(self, rhs):
    return self.sequence_number() > rhs.sequence_number()
  def __le__(self, rhs):
    return self.sequence_number() <= rhs.sequence_number()
  def __ge__(self, rhs):
    return self.sequence_number() >= rhs.sequence_number()
  def __str__(self):
    return self.version
  def __repr__(self):
    return "GrapheneApiVersion('{}')".format(self.version)

  def sequence_number(self):
    return VERSION_MAP[self.version]

  def path(self, graphene_path):
    if self.version == '1.0':
      return self.legacy_path(graphene_path)
    return self.api_vx_path(graphene_path)

  def table_path(self, graphene_path):
    return posixpath.join(graphene_path.modality, 'table', graphene_path.dataset)

  def legacy_path(self, graphene_path):
    """All /segmentation/1.0/$DATASET paths"""
    return posixpath.join(graphene_path.modality, '1.0', graphene_path.dataset)

  def api_vx_path(self, graphene_path):
    """
    All /segmentation/api/v1/$DATASET paths.

    As of Feb. 2020, these were the latest paths.
    """
    return posixpath.join( 
      graphene_path.modality, 'api', self.version, 'table', graphene_path.dataset
    )

class GrapheneMetadata(PrecomputedMetadata):
  def __init__(
    self, cloudpath, use_https=False, 
    use_auth=True, auth_token=None, 
    agglomerate=False,
    *args, **kwargs
  ):
    self.server_url = cloudpath.replace('graphene://', '')
    self.server_path = extract_graphene_path(self.server_url)
    self.use_https = use_https
    self.agglomerate = agglomerate
    self.auth_header = None
    self.spatial_index = None
    if use_auth:
      self.auth_header = {
        "Authorization": "Bearer %s" % self.parse_token(auth_token)
      }
    kwargs['use_https'] = bool(use_https)
    super(GrapheneMetadata, self).__init__(cloudpath, *args, **kwargs)

    version = self.server_path.version
    if version == 'table':
      version = self.supported_api_versions[-1].version

    self.api_version = GrapheneApiVersion(version)

  def parse_token(self, auth_token):
    token = None
    if auth_token:
      token = auth_token
    elif chunkedgraph_credentials:
      token = chunkedgraph_credentials

    if isinstance(token, str):
      try:
        token = json.loads(token)
      except json.decoder.JSONDecodeError:
        pass

    if isinstance(token, dict):
      token = token["token"]

    if token is None:
      raise exceptions.AuthenticationError(
        "No Graphene authentication token was provided. "
        "Does ~/.cloudvolume/secrets/chunkedgraph-secret.json exist?"
      )
    elif not re.match(r'^[0-9a-f]+$', token):
      raise exceptions.AuthenticationError("Graphene authentication token was not formatted correctly. It should be a hexadecimal string.")

    return token

  def supports_api(self, version):
    return GrapheneApiVersion(version) in self.supported_api_versions

  @property  
  def supported_api_versions(self):
    versions = [ 
      GrapheneApiVersion(VERSION_ORDERING[i]) \
      for i in self.info['app']['supported_api_versions'] 
    ]
    versions.sort(key=lambda ver: ver.sequence_number())
    return versions

  @property
  def base_path(self):
    path = self.server_path
    if path.subdomain is None:
      return path.scheme + '://' + path.domain + '/'   
    return path.scheme + '://' + path.subdomain + '.' + path.domain + '/' 

  @property
  def table_path(self):
    return posixpath.join(self.base_path, self.server_path.modality, 'table', self.server_path.dataset)

  @property
  def info_path(self):
    """e.g. https://SUBDOMAIN.dynamicannotationframework.com/segmentation/table/DATASET/info"""
    return posixpath.join(self.table_path, 'info')

  def fetch_info(self):
    """
    Reads info from chunkedgraph endpoint and extracts relevant information
    """
    r = requests.get(self.info_path, headers=self.auth_header)
    r.raise_for_status()
    return r.json()

  @property
  def mesh_path(self):
    if 'mesh' in self.info:
      return self.info['mesh']
    return 'mesh'

  @property
  def cloudpath(self):
    data_dir = self.info['data_dir']
    if self.use_https:
      data_dir = paths.to_https_protocol(data_dir)
    return data_dir

  @property
  def fan_out(self):
    """Number of chunks agglomerated into a new chunk per a level increase in the graph."""
    graph_object = self.info['graph']
    return int(graph_object.get('fan_out', 2))

  def decode_label(self, label):
    level = self.decode_layer_id(label)
    x,y,z = self.decode_chunk_position(label)
    segid = self.decode_segid(label)
    return GrapheneLabel(level, x, y, z, segid)

  def decode_segid(self, label):
    label = uint64(label)
    level = self.decode_layer_id(label)
    segid_bits = self.segid_bits(level)

    mask = uint64(2 ** segid_bits) - uint64(1)
    
    return label & mask

  def decode_chunk_id(self, label):
    """The chunk id is a graphene label with the segid portion zeroed out."""
    label = uint64(label)
    level = self.decode_layer_id(label)
    bits = self.segid_bits(level)
    mask = uint64(2 ** bits) - uint64(1)
    return label & ~mask

  def decode_chunk_position_number(self, label):
    """Returns the chunk position X,Y,Z as a packed uint64."""
    label = uint64(label)
    level = self.decode_layer_id(label)
    ct = self.spatial_bit_count(level)
    label = label & uint64(0x00ffffffffffffff)
    return label >> uint64(self.segid_bits(level))

  def decode_chunk_position(self, label):
    """Returns the chunk position as a tuple (X,Y,Z)"""
    label = uint64(label)
    level = self.decode_layer_id(label)
    ct = self.spatial_bit_count(level)
    label = label & uint64(0x00ffffffffffffff)
    masks = self.spatial_bit_masks(level)
    segid_bits = self.segid_bits(level)

    x = (label & masks[0]) >> uint64(segid_bits + 2 * ct)
    y = (label & masks[1]) >> uint64(segid_bits + 1 * ct)
    z = (label & masks[2]) >> uint64(segid_bits + 0 * ct)

    return Vec(x,y,z)

  def point_to_chunk_position(self, pt, mip=None):
    """
    Convert a point into the chunk position.

    pt: x,y,z triple
    mip: 
      if None, pt is in physical coordinates
      else pt is in the coordinates of the indicated mip level

    Returns: Vec(chunk_x,chunk_y,chunk_z)
    """
    pt = Vec(*pt, dtype=np.float)

    if mip is not None:
      pt *= self.resolution(mip)

    pt /= self.resolution(self.watershed_mip)

    if self.chunks_start_at_voxel_offset:
      pt -= self.voxel_offset(self.watershed_mip)

    return (pt // self.graph_chunk_size).astype(np.int32)

  def point_to_chunk_bbox(self, pt, mip=None):
    """
    For a given point, get the Bbox of the containing 
    chunk.

    pt: x,y,z triple
    mip: 
      if None, pt is in physical coordinates
      else pt is in the coordinates of the indicated mip level

    Returns: Bbox in voxels
    """
    pos = self.point_to_chunk_position(pt, mip)
    pos *= self.graph_chunk_size
    if self.chunks_start_at_voxel_offset:
      pos += self.voxel_offset(self.watershed_mip)

    return Bbox(pos, pos + self.graph_chunk_size)

  def segid_bits(self, level):
    ct = self.spatial_bit_count(level)
    return uint64(64 - self.n_bits_for_layer_id - 3 * ct)

  def decode_layer_id(self, label):
    return uint64(label) >> uint64(64 - self.n_bits_for_layer_id)

  def encode_label(self, layer, x, y, z, segid):
    """
    Create a graphene label from the specified values.

    Another way to use this:

    glabel = GrapheneLabel(2,1,1,1,777)
    meta.encode_label(*glabel)
    """
    if layer > self.n_layers:
      raise ValueError("Provided layer %d is greater than the number of layers in the dataset: %d" % layer, self.n_layers)

    layer_offset = uint64(64 - self.n_bits_for_layer_id)
    bits_per_dim = uint64(self.spatial_bit_count(layer))
    x_offset = uint64(layer_offset - bits_per_dim)
    y_offset = uint64(x_offset - bits_per_dim)
    z_offset = uint64(y_offset - bits_per_dim)

    if not (
      x < 2 ** bits_per_dim and y < 2 ** bits_per_dim and z < 2 ** bits_per_dim
    ):
      raise ValueError(
        "Chunk coordinate is out of range for "
        "this graph on layer %d with %d bits/dim. "
        "[%d, %d, %d]; max = %d."
        % (layer, bits_per_dim, x, y, z, 2 ** bits_per_dim)
      )

    if segid >= 2 ** self.segid_bits(layer):
      raise ValueError(
        "segid {} provided is out of range. It must be less than {}".format(
          segid, 2 ** self.segid_bits(layer)
      ))

    layer = uint64(layer)
    x, y, z = uint64(x), uint64(y), uint64(z)
    segid = uint64(segid)

    return uint64(
      layer << layer_offset | x << x_offset | y << y_offset | z << z_offset | segid
    )

  def spatial_bit_masks(self, level):
    ct = self.spatial_bit_count(level)

    mask = uint64(2 ** ct) - uint64(1)
    segid_bits = 64 - self.n_bits_for_layer_id - 3 * ct

    return [
      mask << uint64(segid_bits + 2 * ct),
      mask << uint64(segid_bits + 1 * ct),
      mask << uint64(segid_bits + 0 * ct)
    ]

  def spatial_bit_count(self, level):
    """
    64-bit labels

    8-bit  chunk coord
    layer | X | Y | Z | segid

    This method returns how many bits in X,Y,Z
    """
    return int(self.info['graph']['spatial_bit_masks'][str(level)])

  @property
  def n_bits_for_layer_id(self):
    return int(self.info['graph'].get('n_bits_for_layer_id', 8))

  @property
  def n_layers(self):
    return int(self.info['graph']['n_layers'])

  @property
  def graph_chunk_size(self):
    return self.info['graph']['chunk_size']

  @property
  def uses_new_draco_bin_size(self):
    graph_object = self.info['graph']
    return int(graph_object.get('uses_new_draco_bin_size', False))
  
  @property
  def mesh_chunk_size(self):
    # TODO: add this as new parameter to the info as it can be different from the chunkedgraph chunksize
    return self.graph_chunk_size

  @property
  def manifest_endpoint(self):
    pth = self.server_path
    pth = GraphenePath(
      pth.scheme, pth.subdomain, pth.domain, 
      'meshing', pth.version, pth.dataset
    )

    url = self.api_version.path(pth)
    return posixpath.join(self.base_path, url, 'manifest')

  @property
  def chunks_start_at_voxel_offset(self):
    """
    Boolean property specifying whether ChunkedGraph chunks begin
    at voxel offset or at origin.
    """
    if 'chunks_start_at_voxel_offset' in self.info:
      return self.info["chunks_start_at_voxel_offset"]
    return False

  @property
  def mesh_metadata(self):
    if 'mesh_metadata' in self.info:
      return self.info["mesh_metadata"]
    return None

  @property
  def uniform_draco_grid_size(self):
    """
    If not None, a number that specifies the draco_grid_size at every ChunkedGraph level.
    """
    if self.mesh_metadata and 'uniform_draco_grid_size' in self.mesh_metadata:
      return self.mesh_metadata["uniform_draco_grid_size"]
    return None

  @property
  def max_meshed_layer(self):
    """
    The highest level in the ChunkedGraph that we create meshes for in this dataset.
    """
    if self.mesh_metadata and 'max_meshed_layer' in self.mesh_metadata:
      return self.mesh_metadata["max_meshed_layer"]
    return None

  @property
  def watershed_mip(self):
    """mip level of the base segmentation that all chunk graph operations remap."""
    return self.info["graph"]["cv_mip"]

  def get_draco_grid_size(self, level):
    """
    Returns the draco_grid_size for specified ChunkedGraph level.
    """
    if self.mesh_metadata is None:
      raise ValueError('This layer is not draco meshed')
    if self.uniform_draco_grid_size is not None:
      return self.uniform_draco_grid_size
    if self.mesh_metadata["max_meshed_layer"] < level:
      raise ValueError(
        "Request level",
        level,
        ". But the maximum meshed level is ",
        self.mesh_metadata["max_meshed_layer"],
      )
    return self.mesh_metadata["draco_grid_sizes"][str(level)]

GraphenePath = namedtuple('GraphenePath', ('scheme', 'subdomain', 'domain', 'modality', 'version', 'dataset'))
LEGACY_EXTRACTION_RE = re.compile(r'/?(\w+)/([\d\.]+)/([\w\d\.\_\-]+)/?')
API_VX_EXTRACTION_RE = re.compile(r'/?(\w+)/api/(v[\d\.]+)/([\w\d\.\_\-]+)/?')
LATEST_API_EXTRACTION_RE = re.compile(r'/?(\w+)/(table)/([\w\d\.\_\-]+)/?')

def extract_graphene_path(url):
  """
  Examples:
  Legacy endpoint:
    graphene://https://SUBDOMAIN.dynamicannotationframework.com/segmentation/1.0/DATASET
  Newer endpoint:
    graphene://https://SUBDOMAIN.dynamicannotationframework.com/segmentation/api/v1/DATASET
  Latest endpoint:
    graphene://https://SUBDOMAIN.DOMAIN_DOT_COM/segmentation/table/DATASET
  """
  parse = urllib.parse.urlparse(url)
  subdomain = parse.netloc.split('.')[0]
  domain = '.'.join(parse.netloc.split('.')[1:])

  schemes = [ 
    LATEST_API_EXTRACTION_RE, API_VX_EXTRACTION_RE, LEGACY_EXTRACTION_RE 
  ]

  for scheme in schemes:
    match = re.match(scheme, parse.path)
    if match:
      break
  else:
    raise exceptions.UnsupportedFormatError("Unable to parse Graphene URL: " + url)

  modality, version, dataset = match.groups()
  return GraphenePath(parse.scheme, subdomain, domain, modality, version, dataset)




import os

from .metadata import PrecomputedMeshMetadata
from .unsharded import UnshardedLegacyPrecomputedMeshSource
from .multilod import ShardedMultiLevelPrecomputedMeshSource, UnshardedMultiLevelPrecomputedMeshSource


from ..metadata import PrecomputedMetadata
from ....cacheservice import CacheService
from ....exceptions import UnsupportedFormatError
from ....lib import red
from ....paths import strict_extract
from ....cloudvolume import SharedConfiguration

class PrecomputedMeshSource(object):
  def __new__(cls, meta, cache, config, readonly=False):
    mesh_meta = PrecomputedMeshMetadata(meta, cache)
    if mesh_meta.info.get('@type', None) == 'neuroglancer_multilod_draco':
      if mesh_meta.info['sharding']['@type'] == 'neuroglancer_uint64_sharded_v1':
        # Sharded storage of multi-resolution mesh fragment data
        return ShardedMultiLevelPrecomputedMeshSource(mesh_meta, cache, config, readonly) 
      else:
        # Unsharded storage of multi-resolution mesh fragment data
        return UnshardedMultiLevelPrecomputedMeshSource()
  
    # Legacy single-resolution mesh format
    return UnshardedLegacyPrecomputedMeshSource(mesh_meta, cache, config, readonly)

  @classmethod
  def from_cloudpath(cls, cloudpath, cache=False, progress=False):
    config = SharedConfiguration(
      cdn_cache=False,
      compress=True,
      compress_level=None,
      green=False,
      mip=0,
      parallel=1,
      progress=progress,
    )

    cache = CacheService(
      cloudpath=(cache if type(cache) == str else cloudpath),
      enabled=bool(cache),
      config=config,
      compress=True,
    )

    cloudpath, mesh_dir = os.path.split(cloudpath)
    meta = PrecomputedMetadata(cloudpath, cache, info={ 'mesh': mesh_dir })

    return PrecomputedMeshSource(meta, cache, config)





from cloudfiles.exceptions import CompressionError, DecompressionError

class AuthenticationError(BaseException):
  """Incorrect credentials."""
  pass

class InfoError(Exception):
  pass

class InfoUnavailableError(InfoError):
  """CloudVolume was unable to access this layer's info file."""
  pass

class RedirectError(InfoError):
  pass

class TooManyRedirects(RedirectError):
  """The chain of redirects became unconvincingly long."""
  pass

class CyclicRedirect(RedirectError):
  """Unable to resolve redirects due to cycle."""
  pass

class ScaleUnavailableError(IndexError):
  """The info file is not configured to support this scale / mip level."""
  pass

class ReadOnlyException(Exception):
  """Attempted to write to a readonly data source."""
  pass

class AlignmentError(ValueError):
  """Signals that an operation requiring chunk alignment was not aligned."""
  pass 

class EmptyVolumeException(Exception):
  """Raised upon finding a missing chunk."""
  pass

class EmptyFileException(Exception):
  """File was zero bytes."""
  pass

class EmptyRequestException(ValueError):
  """
  Requesting uploading or downloading 
  a bounding box of less than one cubic voxel
  is impossible.
  """
  pass

class DecodingError(Exception):
  """Generic decoding error. Applies to content aware and unaware codecs."""
  pass

class EncodingError(Exception):
  """Generic decoding error. Applies to content aware and unaware codecs."""
  pass

class OutOfBoundsError(ValueError):
  """
  Raised upon trying to obtain or assign to a bbox of a volume outside
  of the volume's bounds
  """

class UnsupportedCompressionType(ValueError):
  """
  Raised when attempting to use a compression type which is unsupported
  by the storage interface.
  """

# Inheritance below done for backwards compatibility reasons.

class MeshDecodeError(ValueError):
  """Unable to decode a mesh object."""
  pass

class SkeletonUnassignedEdgeError(Exception):
  """This skeleton has an edge to a vertex that doesn't exist."""
  pass

class SkeletonDecodeError(Exception):
  """Unable to decode a binary skeleton into a Python object."""
  pass

class SkeletonEncodeError(Exception):
  """Unable to encode a PrecomputedSkeleton into a binary object."""
  pass

class SkeletonTransformError(Exception):
  """Unable to apply a spatial transfrom to the current coordinate system."""
  pass

class SkeletonAttributeMixingError(Exception):
  """
  These skeletons have different vertex attributes 
  and cannot be recombined without manual intervention.
  """
  pass

class UnsupportedFormatError(Exception):
  """Unable to interpret the format of this URI. e.g. precomputed://"""
  pass

class UnsupportedProtocolError(ValueError):
  """Unknown protocol extension."""
  pass

class UnsupportedGrapheneAPIVersionError(Exception):
  """This dataset does not support the specified api version."""
  pass

class SpecViolation(Exception):
  """The values held by this object violate its written specification."""
  pass

class SpatialIndexGapError(Exception):
  """Part of the spatial index was not found. A complete result set cannot be fetched."""
  pass

class WriteLockAcquisitionError(Exception):
  """Unable to obtain a lock on this data layer element."""
  pass

class WriteLockReleaseError(Exception):
  """Unable to release a lock on this data layer element."""
  pass


from __future__ import print_function

from intern.remote.boss import BossRemote
from intern.resource.boss.resource import (
  ChannelResource, ExperimentResource, CoordinateFrameResource
)

from .. import autocropfn, readonlyguard, ImageSourceInterface

from ... import exceptions 
from ...lib import ( 
  colorize, red, mkdir, Vec, Bbox,  
  jsonify, generate_random_string
)
from ...secrets import boss_credentials
from ...volumecutout import VolumeCutout

class BossImageSource(ImageSourceInterface):
  def __init__(
    self, config, meta, cache,
    autocrop=False, bounded=True,
    non_aligned_writes=False,
    delete_black_uploads=False,
    readonly=False,
  ):
    self.config = config
    self.meta = meta 
    self.cache = cache 

    self.autocrop = bool(autocrop)
    self.bounded = bool(bounded)
    self.non_aligned_writes = bool(non_aligned_writes)
    self.readonly = bool(readonly)

  def download(self, bbox, mip, parallel=1, renumber=False):
    if parallel != 1:
      raise ValueError("Only parallel=1 is supported for boss.")
    elif renumber != False:
      raise ValueError("Only renumber=False is supported for boss.")

    bounds = Bbox.clamp(bbox, self.meta.bounds(mip))

    if self.autocrop:
      image, bounds = autocropfn(self.meta, image, bounds, mip)
    
    if bounds.subvoxel():
      raise exceptions.EmptyRequestException('Requested less than one pixel of volume. {}'.format(bounds))

    x_rng = [ bounds.minpt.x, bounds.maxpt.x ]
    y_rng = [ bounds.minpt.y, bounds.maxpt.y ]
    z_rng = [ bounds.minpt.z, bounds.maxpt.z ]

    layer_type = 'image' if self.meta.layer_type == 'unknown' else self.meta.layer_type

    chan = ChannelResource(
      collection_name=self.meta.path.bucket, 
      experiment_name=self.meta.path.dataset, 
      name=self.meta.path.layer, # Channel
      type=layer_type, 
      datatype=self.meta.data_type,
    )

    rmt = BossRemote(boss_credentials)
    cutout = rmt.get_cutout(chan, mip, x_rng, y_rng, z_rng, no_cache=True)
    cutout = cutout.T
    cutout = cutout.astype(self.meta.dtype)
    cutout = cutout[::steps.x, ::steps.y, ::steps.z]

    if len(cutout.shape) == 3:
      cutout = cutout.reshape(tuple(list(cutout.shape) + [ 1 ]))

    if self.bounded or self.autocrop or bounds == bbox:
      return VolumeCutout.from_volume(self.meta, mip, cutout, bounds)

    # This section below covers the case where the requested volume is bigger
    # than the dataset volume and the bounds guards have been switched 
    # off. This is useful for Marching Cubes where a 1px excess boundary
    # is needed.
    shape = list(bbox.size3()) + [ cutout.shape[3] ]
    renderbuffer = np.zeros(shape=shape, dtype=self.meta.dtype, order='F')
    txrx.shade(renderbuffer, bbox, cutout, bounds)
    return VolumeCutout.from_volume(self.meta, mip, renderbuffer, bbox)

  @readonlyguard
  def upload(self, image, offset, mip):
    shape = Vec(*image.shape[:3])
    offset = Vec(*offset)

    bounds = Bbox(offset, shape + offset)

    if bounds.subvoxel():
      raise exceptions.EmptyRequestException('Requested less than one pixel of volume. {}'.format(bounds))

    if self.autocrop:
      image, bounds = autocropfn(self.meta, image, bounds, mip)
      offset = bounds.minpt

    check_grid_aligned(
      self.meta, image, bounds, mip, 
      throw_error=(self.non_aligned_writes == False)
    )

    x_rng = [ bounds.minpt.x, bounds.maxpt.x ]
    y_rng = [ bounds.minpt.y, bounds.maxpt.y ]
    z_rng = [ bounds.minpt.z, bounds.maxpt.z ]

    layer_type = 'image' if self.layer_type == 'unknown' else self.meta.layer_type

    chan = ChannelResource(
      collection_name=self.meta.path.bucket, 
      experiment_name=self.meta.path.dataset, 
      name=self.meta.path.layer, # Channel
      type=layer_type, 
      datatype=self.meta.data_type,
    )

    if image.shape[3] == 1:
      image = image.reshape( image.shape[:3] )

    rmt = BossRemote(boss_credentials)
    image = image.T
    image = np.asfortranarray(image.astype(self.meta.dtype))

    rmt.create_cutout(chan, mip, x_rng, y_rng, z_rng, image)

  def exists(self, bbox, mip=None):
    raise NotImplementedError()

  @readonlyguard
  def delete(self, bbox, mip=None):
    raise NotImplementedError()

  def transfer_to(self, cloudpath, bbox, mip, block_size=None, compress=True):
    raise NotImplementedError()